/*--------------------------------------------------------------------------------*/
/* WINDOW VOLUME CONTROL FOR ONE APPLICATION                      V.Burel (c)2018 */
/*                                                                                */
/* SOURCE CODE EXAMPLE FOR VALIDATION TEST ONLY                                   */
/*                                                                                */
/* WEB  : www.vb-audio.com                                                        */
/*--------------------------------------------------------------------------------*/
/*  Minimal WIN32 program showing how an AUDIO THREAD can be pre-empted/disturbed */
/*  under WIN10 for a big amount of time (> 5 ms) in a simple memory copy         */
/*  operation. (to be compared to WIN7 Real Time audio capabilities)              */
/*--------------------------------------------------------------------------------*/
/*  To compile With Microsoft VC2005 to 2017 (x86 or x64 target).                 */
/*  Create WIN32 empty project with the following options :                       */
/*  - Configuration Properties / General : Char Set = NOT SET                     */
/*  - Configuration Properties / C/C++ / Preprocessor : _CRT_SECURE_NO_WARNINGS   */
/*  LINKER: Input / Additional Dependencies: -                                    */
/*--------------------------------------------------------------------------------*/
/* THIS SOURCE CODE CAN BE USED / MODIFIED and PUBLISHED TO MAKE FURTHER TEST     */
/*--------------------------------------------------------------------------------*/

#ifndef __cplusplus
	#ifndef STRICT
		#define STRICT
	#endif
#endif

#include <windows.h>
#include <stdio.h>

#include <MMDeviceAPI.h>
#include <AudioPolicy.h>


//
//define globals 
//

#define SZAPPPUBLICNAME "ISimpleAudioVolume::SetMasterVolume TEST"

#define UI_WIN_DX	800						// window width
#define UI_WIN_DY	400						// window height

#define TIMER_ID	16325

#define IDB_SWAP_MUTE		100
#define IDB_SETLEVEL_0		110
#define IDB_SETLEVEL_25		111
#define IDB_SETLEVEL_50		112
#define IDB_SETLEVEL_100	113

//
// APP OBJECT
//

typedef struct tagAPPCTX
{
	HWND hwnd_MainWindow;
	HINSTANCE hinstance;

	UINT_PTR	display_Timer;
	RECT		rect_volmute;
	HWND		button_Mute;
	HWND		button_0;
	HWND		button_25;
	HWND		button_50;
	HWND		button_100;


	IMMDevice				*pAudioDevice;
	IAudioSessionManager2	*pAudioSessionManager2;
    IAudioSessionControl	*pSessionControl;
	IAudioSessionEvents		iAudioSessionEvents;
	
	ULONG	cRefCounter;

	WCHAR		current_session_wsz[512];
	float		current_volume;
	int			current_mute;
	int			fAppDisconnect; //for debugging only

	float		displayed_volume;
	int			displayed_mute;


} T_APPCTX, *PT_APPCTX, *LPT_APPCTX;

static T_APPCTX G_appctx = {NULL, NULL};

/*******************************************************************************/
/**                              REQUIRED GUID                                **/
/*******************************************************************************/


	//class DECLSPEC_UUID("BCDE0395-E52F-467C-8E3D-C4579291692E")
	static GUID localCLSID_MMDeviceEnumerator={0xBCDE0395,0xE52F,0x467C,0x8E, 0x3D, 0xC4,0x57,0x92,0x91,0x69,0x2E};

    //MIDL_INTERFACE("A95664D2-9614-4F35-A746-DE8DB63617E6")
	static GUID localIID_IMMDeviceEnumerator={0xA95664D2,0x9614,0x4F35,0xA7,0x46,0xDE,0x8D,0xB6,0x36,0x17,0xE6};

    //MIDL_INTERFACE("77AA99A0-1BD6-484F-8BC7-2C654C9A9B6F")
	static GUID localIID_IAudioSessionManager2={0x77AA99A0,0x1BD6,0x484F,0x8B,0xC7,0x2C,0x65,0x4C,0x9A,0x9B,0x6F};

	//MIDL_INTERFACE("bfb7ff88-7239-4fc9-8fa2-07c950be9c6d")
	static GUID localIID_IAudioSessionControl2={0xbfb7ff88, 0x7239, 0x4fc9, 0x8f,0xa2, 0x07,0xc9,0x50,0xbe,0x9c,0x6d};

	//MIDL_INTERFACE("87ce5498-68d6-44e5-9215-6da47ef883d8")
	static GUID localIID_ISimpleAudioVolume={0x87ce5498, 0x68d6, 0x44e5, 0x92, 0x15, 0x6d, 0xa4, 0x7e, 0xf8, 0x83, 0xd8};

	//IID_IAudioSessionEvents "24918ACC-64B3-37C1-8CA9-74A66E9957A8"
	static GUID localIID_IAudioSessionEvents={0x24918ACC, 0x64B3, 0x37C1, 0x8C, 0xA9, 0x74, 0xA6, 0x6E, 0x99, 0x57, 0xA8};


	//GUID.Empty
	static GUID localIID_NULL={0x00000000, 0x0000, 0x0000, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00};
	static GUID localIID_MyOwnGUID={0x5c392165, 0xf5c, 0x457b, 0x88, 0x40, 0x22, 0xab, 0xf2, 0x1b, 0xe8, 0x57};


/*******************************************************************************/
/**              IAudioSessionEvents Interface / C-Style Object               **/
/*******************************************************************************/
/* we build our IAudioSessionEvents interface to receive Audio Session Control */
/* notifications, especially  _OnSimpleVolumeChanged                           */
/*******************************************************************************/

#define _OFFSET(_Type,_Member) (UINT_PTR)&(((_Type *)0)->_Member)

#define _GETCLASSADDRESS(_classType, _member, _pointer)\
	(&((_classType*)0)->_member == _pointer, ((_classType *)(((UINT_PTR)_pointer)-_OFFSET(_classType,_member))))

#define _VTBL_OF(_pointer)	((IUnknownVtbl *)(((IUnknown *)(_pointer))->lpVtbl))

#define _ADDREF(pif)	(_VTBL_OF(pif)->AddRef((IUnknown *)(pif)))
#define _RELEASE(pif) (_VTBL_OF(pif)->Release((IUnknown *)(pif)))

//
//                          IUNKNONWN
//

static HRESULT __stdcall iAudioSessionEvents_QueryInterface(IAudioSessionEvents * pfi, REFIID riid, void **ppv)
{
	LPT_APPCTX lpo;
	if (pfi == NULL) return E_FAIL;
	lpo=_GETCLASSADDRESS(T_APPCTX, iAudioSessionEvents, pfi);
	if (ppv == NULL) return E_FAIL;
	*ppv=NULL;
	//-----------------------------
	if ((IsEqualIID(riid, &IID_IUnknown) == TRUE) ||
		(IsEqualIID(riid, &localIID_IAudioSessionEvents) == TRUE))
		//(IsEqualIID(riid, &IID_IDispatch)==TRUE ))
	{
		*ppv=&(lpo->iAudioSessionEvents);
		_ADDREF(*ppv);
		return NOERROR;
	}
	return E_NOINTERFACE;
}

static ULONG __stdcall iAudioSessionEvents_AddRef(IAudioSessionEvents * pfi)
{
	LPT_APPCTX lpo;
	if (pfi == NULL) return (ULONG)E_FAIL;
	lpo=_GETCLASSADDRESS(T_APPCTX, iAudioSessionEvents, pfi);
	lpo->cRefCounter++;
	return lpo->cRefCounter;
}

static ULONG __stdcall iAudioSessionEvents_Release(IAudioSessionEvents * pfi)
{
	LPT_APPCTX lpo;
	if (pfi == NULL) return (ULONG)E_FAIL;
	lpo=_GETCLASSADDRESS(T_APPCTX, iAudioSessionEvents, pfi);
	lpo->cRefCounter--;
	if (lpo->cRefCounter < 1)
	{
		lpo->cRefCounter=0;
		return 0;	
	}
	return lpo->cRefCounter;
}

//
//                          OTHER METHODS
//

       
static HRESULT __stdcall iAudioSessionEvents_OnDisplayNameChanged(IAudioSessionEvents * pfi, 
            LPCWSTR NewDisplayName,
            LPCGUID EventContext)
{
	LPT_APPCTX lpo;
	if (pfi == NULL) return E_FAIL;
	lpo=_GETCLASSADDRESS(T_APPCTX, iAudioSessionEvents, pfi);
	return S_OK;
}
        
static HRESULT __stdcall iAudioSessionEvents_OnIconPathChanged(IAudioSessionEvents * pfi,  
            LPCWSTR NewIconPath,
            LPCGUID EventContext)
{
	LPT_APPCTX lpo;
	if (pfi == NULL) return E_FAIL;
	lpo=_GETCLASSADDRESS(T_APPCTX, iAudioSessionEvents, pfi);
	return S_OK;
}
        
static HRESULT __stdcall iAudioSessionEvents_OnSimpleVolumeChanged(IAudioSessionEvents * pfi,  
            float NewVolume,
            BOOL NewMute,
            LPCGUID EventContext)
{
	LPT_APPCTX lpo;
	if (pfi == NULL) return E_FAIL;
	lpo=_GETCLASSADDRESS(T_APPCTX, iAudioSessionEvents, pfi);
	lpo->current_volume = NewVolume;
	lpo->current_mute = (int)NewMute;
	return S_OK;
}
        
static HRESULT __stdcall iAudioSessionEvents_OnChannelVolumeChanged(IAudioSessionEvents * pfi,  
            DWORD ChannelCount,
            float NewChannelVolumeArray[],
            DWORD ChangedChannel,
            LPCGUID EventContext)
{
	LPT_APPCTX lpo;
	if (pfi == NULL) return E_FAIL;
	lpo=_GETCLASSADDRESS(T_APPCTX, iAudioSessionEvents, pfi);
	return S_OK;
}
        
static HRESULT __stdcall iAudioSessionEvents_OnGroupingParamChanged(IAudioSessionEvents * pfi,  
            LPCGUID NewGroupingParam,
            LPCGUID EventContext)
{
	LPT_APPCTX lpo;
	if (pfi == NULL) return E_FAIL;
	lpo=_GETCLASSADDRESS(T_APPCTX, iAudioSessionEvents, pfi);
	return E_NOTIMPL;
}
        
static HRESULT __stdcall iAudioSessionEvents_OnStateChanged(IAudioSessionEvents * pfi,  
            AudioSessionState NewState)
{
	LPT_APPCTX lpo;
	if (pfi == NULL) return E_FAIL;
	lpo=_GETCLASSADDRESS(T_APPCTX, iAudioSessionEvents, pfi);
	switch(NewState)
	{
	case AudioSessionStateActive:
		lpo->fAppDisconnect=0;
		break;
	case AudioSessionStateInactive:
		lpo->fAppDisconnect=0;
		break;
	case AudioSessionStateExpired:	//this is not working with Media Player Close
		lpo->fAppDisconnect=1;
		break;
	}
	return S_OK;
}
        
static HRESULT __stdcall iAudioSessionEvents_OnSessionDisconnected(IAudioSessionEvents * pfi,  
            AudioSessionDisconnectReason DisconnectReason)
{
	LPT_APPCTX lpo;
	if (pfi == NULL) return E_FAIL;
	lpo=_GETCLASSADDRESS(T_APPCTX, iAudioSessionEvents, pfi);
	lpo->fAppDisconnect=1;
	return S_OK;
}


static IAudioSessionEventsVtbl vtblIAudioSessionEvents=
{
	iAudioSessionEvents_QueryInterface,
	iAudioSessionEvents_AddRef,
	iAudioSessionEvents_Release,
	//--------------------------
	iAudioSessionEvents_OnDisplayNameChanged,
	iAudioSessionEvents_OnIconPathChanged,
	iAudioSessionEvents_OnSimpleVolumeChanged,
	iAudioSessionEvents_OnChannelVolumeChanged,
	iAudioSessionEvents_OnGroupingParamChanged,
	iAudioSessionEvents_OnStateChanged,
	iAudioSessionEvents_OnSessionDisconnected
};





/*******************************************************************************/
/**                     SET AudioSession Volume & Mute                        **/
/*******************************************************************************/


void VOLUMECTRL_ConnectedApp_SetVolume(LPT_APPCTX lpapp, float flevel)
{
	HRESULT hr;
	ISimpleAudioVolume* pSimpleAudioVolume=NULL;
	if (lpapp->pSessionControl == NULL) return;

	hr=lpapp->pSessionControl->lpVtbl->QueryInterface(lpapp->pSessionControl, &localIID_ISimpleAudioVolume, (void**)(&pSimpleAudioVolume));
	if (pSimpleAudioVolume != NULL)
	{
		//hr=pSimpleAudioVolume->lpVtbl->SetMasterVolume(pSimpleAudioVolume, flevel, &localIID_MyOwnGUID);
		hr=pSimpleAudioVolume->lpVtbl->SetMasterVolume(pSimpleAudioVolume, flevel, &localIID_NULL);
		pSimpleAudioVolume->lpVtbl->Release(pSimpleAudioVolume);
	}
}

void VOLUMECTRL_ConnectedApp_SetMute(LPT_APPCTX lpapp, BOOL fMute)
{
	HRESULT hr;
	ISimpleAudioVolume* pSimpleAudioVolume=NULL;
	if (lpapp->pSessionControl == NULL) return;

	hr=lpapp->pSessionControl->lpVtbl->QueryInterface(lpapp->pSessionControl, &localIID_ISimpleAudioVolume, (void**)(&pSimpleAudioVolume));
	if (pSimpleAudioVolume != NULL)
	{
		//hr=pSimpleAudioVolume->lpVtbl->SetMute(pSimpleAudioVolume, fMute, &localIID_MyOwnGUID);
		hr=pSimpleAudioVolume->lpVtbl->SetMute(pSimpleAudioVolume, fMute, &localIID_NULL);
		pSimpleAudioVolume->lpVtbl->Release(pSimpleAudioVolume);
	}
}


/*******************************************************************************/
/**            GET AudioSession On Default Playack device                     **/
/*******************************************************************************/

void VOLUMECTRL_EnumerateFirstConnectedApplication(LPT_APPCTX lpapp)
{
	int nbSession,vi;
	HRESULT hr, fsystemsound;
	LPWSTR pwchar;
	DWORD processUID;
	BOOL bbb;
    IAudioSessionControl	*pSessionControl=NULL;
    IAudioSessionEnumerator	*pSessionEnumerator=NULL;
	IAudioSessionControl2*	pSessionControl2=NULL;
	ISimpleAudioVolume*		pSimpleAudioVolume=NULL;

	hr = lpapp->pAudioSessionManager2->lpVtbl->GetSessionEnumerator(lpapp->pAudioSessionManager2, (void*)&(pSessionEnumerator));
    if (hr != S_OK) return;
    hr = pSessionEnumerator->lpVtbl->GetCount(pSessionEnumerator, &nbSession);
	if (hr != S_OK) return;

	lpapp->current_session_wsz[0]=0;
    for (vi = 0; vi < nbSession; vi++)
    {
		pSessionControl=NULL;
		hr=pSessionEnumerator->lpVtbl->GetSession(pSessionEnumerator, vi, &pSessionControl);
		if (pSessionControl != NULL)
		{
			pSessionControl2=NULL;
			// Get SessionControl2 Interface to get an identification string of the connected application
			hr=pSessionControl->lpVtbl->QueryInterface(pSessionControl, &localIID_IAudioSessionControl2, (void**)(&pSessionControl2));
			if ((hr == S_OK) && (pSessionControl2 != NULL))
			{
				//we don't want the System Session
				fsystemsound = pSessionControl2->lpVtbl->IsSystemSoundsSession(pSessionControl2);
				if (fsystemsound != S_OK)
				{
					//it must have a process ID
					hr=pSessionControl2->lpVtbl->GetProcessId(pSessionControl2, &processUID);
					if (hr == S_OK)
					{
						//We get Session Identifier string for this connected application
						hr= pSessionControl2->lpVtbl->GetSessionIdentifier(pSessionControl2, &pwchar);
						if (hr == S_OK)
						{
							wcsncpy(lpapp->current_session_wsz, pwchar,511);
							lpapp->current_session_wsz[511]=0;
							CoTaskMemFree(pwchar);
						}
					}
				}
				pSessionControl2->lpVtbl->Release(pSessionControl2);
			}

			//if we found a connected application
			if (lpapp->current_session_wsz[0] != 0)
			{
				// Get Current Volume + Mute State
				hr=pSessionControl->lpVtbl->QueryInterface(pSessionControl, &localIID_ISimpleAudioVolume, (void**)(&pSimpleAudioVolume));
				if (pSimpleAudioVolume != NULL)
				{
					hr=pSimpleAudioVolume->lpVtbl->GetMasterVolume(pSimpleAudioVolume, &(lpapp->current_volume));
					hr=pSimpleAudioVolume->lpVtbl->GetMute(pSimpleAudioVolume, &bbb);
					if (bbb == TRUE) lpapp->current_mute=1;
					else lpapp->current_mute=0;
					pSimpleAudioVolume->lpVtbl->Release(pSimpleAudioVolume);
				}
				// Register IAudioSessionEvents and Stop enumerating Audio Sessions
				lpapp->iAudioSessionEvents.lpVtbl=&vtblIAudioSessionEvents;
				lpapp->cRefCounter=1;
				pSessionControl->lpVtbl->RegisterAudioSessionNotification(pSessionControl, &(lpapp->iAudioSessionEvents));
				lpapp->pSessionControl =pSessionControl;
				goto EndOfProcPoint;
			}
			pSessionControl->lpVtbl->Release(pSessionControl);
		}

    }
EndOfProcPoint:
	if (pSessionEnumerator != NULL) pSessionEnumerator->lpVtbl->Release(pSessionEnumerator);
}


void VOLUMECTRL_GetAudioSessionOnDefaultPlaybackDevice(LPT_APPCTX lpapp)
{
	HRESULT hr;
	IMMDeviceEnumerator *deviceEnumerator = NULL;

	hr= CoCreateInstance(&localCLSID_MMDeviceEnumerator,NULL, CLSCTX_INPROC_SERVER | CLSCTX_LOCAL_SERVER, &localIID_IMMDeviceEnumerator,&deviceEnumerator);
	if (hr == S_OK)
	{
		lpapp->pAudioDevice=NULL;
		hr= deviceEnumerator->lpVtbl->GetDefaultAudioEndpoint(deviceEnumerator, eRender, eMultimedia, &(lpapp->pAudioDevice));
		if (hr == S_OK)
		{
			hr = lpapp->pAudioDevice->lpVtbl->Activate(lpapp->pAudioDevice, &localIID_IAudioSessionManager2, 0, NULL, (void**)&(lpapp->pAudioSessionManager2));
			if (hr == S_OK)
			{
				VOLUMECTRL_EnumerateFirstConnectedApplication(lpapp);
			}
		}
	}
	if (deviceEnumerator != NULL) deviceEnumerator->lpVtbl->Release(deviceEnumerator);
}

void VOLUMECTRL_CleanUpAllInterfaces(LPT_APPCTX lpapp)
{
	if (lpapp->pSessionControl != NULL) 
	{
		lpapp->pSessionControl->lpVtbl->UnregisterAudioSessionNotification(lpapp->pSessionControl, &(lpapp->iAudioSessionEvents));
		lpapp->pSessionControl->lpVtbl->Release(lpapp->pSessionControl);
		lpapp->pSessionControl=NULL;
	}

	if (lpapp->pAudioSessionManager2 != NULL) 
	{
		lpapp->pAudioSessionManager2->lpVtbl->Release(lpapp->pAudioSessionManager2);
		lpapp->pAudioSessionManager2=NULL;
	}
	if (lpapp->pAudioDevice != NULL) 
	{
		lpapp->pAudioDevice->lpVtbl->Release(lpapp->pAudioDevice);
		lpapp->pAudioDevice=NULL;
	}
}










/*******************************************************************************/
/**                                 PAINTING                                  **/
/*******************************************************************************/

void DrawAppVolumeAndMuteState(LPT_APPCTX lpapp, HWND hw, HDC dc)	
{
	char sss[512];
	RECT rect;
	HPEN oldpen;
	HBRUSH oldbrush;

	oldpen=(HPEN)SelectObject(dc,GetStockObject(BLACK_PEN));
	oldbrush=(HBRUSH)SelectObject(dc,GetStockObject(WHITE_BRUSH));
	SetBkMode(dc,TRANSPARENT);

	rect = lpapp->rect_volmute;
	rect.right = rect.left+300;

	RoundRect(dc,rect.left,rect.top,rect.right,rect.bottom,7,7);
	sprintf(sss,"Volume: %0.3f",lpapp->current_volume);
	DrawText(dc, sss,(int)strlen(sss),&rect,DT_SINGLELINE | DT_CENTER | DT_VCENTER);

	rect = lpapp->rect_volmute;
	rect.left +=300+10;
	rect.right = rect.left+150;
	RoundRect(dc,rect.left,rect.top,rect.right,rect.bottom,7,7);
	sprintf(sss,"Mute: %i",lpapp->current_mute);
	DrawText(dc, sss,(int)strlen(sss),&rect,DT_SINGLELINE | DT_CENTER | DT_VCENTER);
	
	SelectObject(dc,oldpen);
	SelectObject(dc,oldbrush);

	lpapp->displayed_volume =lpapp->current_volume;
	lpapp->displayed_mute	=lpapp->current_mute;

}

void DrawAllStuff(LPT_APPCTX lpapp, HWND hw, HDC dc)	
{
	char sss[512];
	RECT rect;
	HPEN oldpen;
	HBRUSH oldbrush;

	oldpen=(HPEN)SelectObject(dc,GetStockObject(WHITE_PEN));
	oldbrush=(HBRUSH)SelectObject(dc,GetStockObject(WHITE_BRUSH));
	SetBkMode(dc,TRANSPARENT);
	
	GetClientRect(hw, &rect);
	rect.left+=10;
	rect.top+=10;
	rect.right-=10;
	rect.bottom=rect.top+20;

	SetTextColor(dc,RGB(80,80,80));
	strcpy(sss,"Current Connected Audio Session Identifier:");
	DrawText(dc, sss,(int)strlen(sss),&rect,DT_SINGLELINE | DT_LEFT | DT_VCENTER);

	SetTextColor(dc,RGB(0,0,0));
	rect.top=rect.bottom+5;
	rect.bottom=rect.top+50;
	RoundRect(dc,rect.left-4,rect.top-4,rect.right+4,rect.bottom+4,7,7);
	DrawTextW(dc, lpapp->current_session_wsz,(int)wcslen(lpapp->current_session_wsz),&rect,DT_WORDBREAK | DT_LEFT | DT_TOP);

	rect.top=rect.bottom+40;
	rect.bottom=rect.top+30;
	lpapp->rect_volmute= rect;

	rect.top-=25;
	rect.bottom=rect.top+20;
	SetTextColor(dc,RGB(80,80,80));
	strcpy(sss,"Current Volume Level a Mute State:");
	DrawText(dc, sss,(int)strlen(sss),&rect,DT_SINGLELINE | DT_LEFT | DT_VCENTER);

	SelectObject(dc,oldpen);
	SelectObject(dc,oldbrush);

	DrawAppVolumeAndMuteState(lpapp, hw, dc);	
}



/*******************************************************************************/
/**                              GLOBAL INIT / END                            **/
/*******************************************************************************/

BOOL InitSoftware(LPT_APPCTX lpapp, HWND hw)
{
	lpapp->button_0=CreateWindow("BUTTON", "0", WS_CHILD | WS_VISIBLE,
									10, 160, 60, 30,
									hw, (HMENU)IDB_SETLEVEL_0, lpapp->hinstance, NULL);
	lpapp->button_25=CreateWindow("BUTTON", "0.25", WS_CHILD | WS_VISIBLE,
									80, 160, 60, 30,
									hw, (HMENU)IDB_SETLEVEL_25, lpapp->hinstance, NULL);
	lpapp->button_50=CreateWindow("BUTTON", "0.50", WS_CHILD | WS_VISIBLE,
									150, 160, 60, 30,
									hw, (HMENU)IDB_SETLEVEL_50, lpapp->hinstance, NULL);
	lpapp->button_100=CreateWindow("BUTTON", "1.00", WS_CHILD | WS_VISIBLE,
									220, 160, 60, 30,
									hw, (HMENU)IDB_SETLEVEL_100, lpapp->hinstance, NULL);

	lpapp->button_Mute=CreateWindow("BUTTON", "Swap Mute", WS_CHILD | WS_VISIBLE,
									320, 160, 150, 30,
									hw, (HMENU)IDB_SWAP_MUTE, lpapp->hinstance, NULL);

	VOLUMECTRL_GetAudioSessionOnDefaultPlaybackDevice(lpapp);
	return TRUE;
}

BOOL EndSoftware(LPT_APPCTX lpapp)
{
	
	VOLUMECTRL_CleanUpAllInterfaces(lpapp);
	return TRUE;
}

/*******************************************************************************/
/*                          WINDOWS MAIN CALLBACK                              */
/*******************************************************************************/

LRESULT CALLBACK MainWindowManageEvent(HWND hw,			//handle of the window.
											UINT msg,   //Message Ident.
											WPARAM p1,	//parameter 1.
											LPARAM p2)	//parameter 2
{
	int nCommand;
	HDC dc;
	PAINTSTRUCT ps;
	LPT_APPCTX lpapp=&G_appctx;
	switch (msg)
	{


		case WM_CREATE:
			if (InitSoftware(lpapp, hw) == FALSE) return -1;//return -1 here cancel the window creation
			lpapp->display_Timer = SetTimer(hw,TIMER_ID, 30,NULL);
			break;
		case WM_COMMAND:
			nCommand = LOWORD(p1);
			switch(nCommand)
			{
			//Note we are not updating our own internal volume/mute values because 
			//the method _OnSimpleVolumeChanged is called by the system after SetMute or SetVolume. 
			case IDB_SWAP_MUTE:
				if (lpapp->current_mute == 0) VOLUMECTRL_ConnectedApp_SetMute(lpapp, TRUE);
				else VOLUMECTRL_ConnectedApp_SetMute(lpapp, FALSE);
				break;
			case IDB_SETLEVEL_0:
				VOLUMECTRL_ConnectedApp_SetVolume(lpapp, 0.0f);
				break;
			case IDB_SETLEVEL_25:
				VOLUMECTRL_ConnectedApp_SetVolume(lpapp, 0.25f);
				break;
			case IDB_SETLEVEL_50:
				VOLUMECTRL_ConnectedApp_SetVolume(lpapp, 0.50f);
				break;
			case IDB_SETLEVEL_100:
				VOLUMECTRL_ConnectedApp_SetVolume(lpapp, 1.0f);
				break;
			}
			break;
		case WM_TIMER:
			// Timer is updating volume or mute state display on value change.
			if (p1 == TIMER_ID)
			{
				if ((lpapp->current_mute != lpapp->displayed_mute) ||
					(lpapp->current_volume != lpapp->displayed_volume))
				{
					dc=GetDC(hw);
					DrawAppVolumeAndMuteState(lpapp, hw, dc);	
					ReleaseDC(hw,dc);
				}
			}
			break;
		case WM_PAINT:
			dc=BeginPaint(hw,&ps);
			DrawAllStuff(lpapp, hw, dc);	
			EndPaint(hw,&ps);
	        break;
		case WM_CLOSE:
			PostMessage(hw,WM_DESTROY,0,0L);
			break;
		case WM_DESTROY:
			if (lpapp->display_Timer != 0) KillTimer(hw,lpapp->display_Timer);

			EndSoftware(lpapp);
			PostQuitMessage(0);
			break;
		default:
			return (DefWindowProc(hw,msg,p1,p2));

	}
	return (0L);
}


/*******************************************************************************/
/**                              MAIN PROCDURE                                **/
/*******************************************************************************/

int APIENTRY WinMain(HINSTANCE handle_app,			//Application hinstance.
							HINSTANCE handle_prev,  //NULL.
							LPTSTR param,           //Command Line Parameter.
							int com_show)           //How to display window (optionnal).
{
	LPT_APPCTX lpapp;
	long	wstyle;
	MSG		msg;    
	char	szWindowClassName[]="MainWindowClass";
	char *	title="Sorry";

	LPVOID lp_coinit=NULL;
	HRESULT hr_coinit;

	WNDCLASS	wc;
	//Init app ctx
	lpapp = &G_appctx;
	memset(lpapp,0,sizeof(T_APPCTX));
	lpapp->hinstance=handle_app;

	//initialize COM stuff
	hr_coinit=CoInitialize(lp_coinit);
	//hr_coinit=CoInitializeEx(NULL, COINIT_MULTITHREADED);

	//we define a window class to create a window from this class 
	wc.style		=CS_HREDRAW | CS_VREDRAW;  	  		//property.
	wc.lpfnWndProc=(WNDPROC)MainWindowManageEvent;		//Adresse of our Callback.
	wc.cbClsExtra =0;					  				//Possibility to store some byte inside a class object.
	wc.cbWndExtra =0;                          			//Possibility to store some byte inside a window object.
	wc.hInstance  =handle_app; 	                		//handle of the application hinstance.
	wc.hIcon      =LoadIcon(NULL, IDI_APPLICATION);    	//handle of icon displayed in the caption.
	wc.hCursor    =LoadCursor(NULL,IDC_ARROW);			//handle of cursor mouse .
	wc.hbrBackground=(HBRUSH)(COLOR_MENU+1);			//Background color.
	wc.lpszMenuName=NULL;			    				//pointer on menu defined in resource.
	wc.lpszClassName=szWindowClassName;       			//pointer on class name.

	//register class.
	if (RegisterClass(&wc)==0)
	{
		MessageBox(NULL,"Failed to register main class...",title,MB_APPLMODAL | MB_OK | MB_ICONERROR);
		return 0;
	}
	//create a windows from this class.
	wstyle=WS_OVERLAPPEDWINDOW | WS_VISIBLE;
	lpapp->hwnd_MainWindow=CreateWindow(szWindowClassName,	// address of registered class name.
						 SZAPPPUBLICNAME,					// address of window name string
						 wstyle,						// window style
						 CW_USEDEFAULT,					// horizontal position of window
						 CW_USEDEFAULT,					// vertical position of window
						 UI_WIN_DX,						// window width
						 UI_WIN_DY,						// window height
						 NULL,							// parent handle is NULL since it's a main window.
						 NULL,							// menu name defined in resource (NULL if no menu or already defined in the Class).
						 handle_app,					// handle of application instance
						 NULL); 						// address of window-creation data

	if (lpapp->hwnd_MainWindow==NULL)
	{
		MessageBox(NULL,"Failed to create window...",title,MB_APPLMODAL | MB_OK | MB_ICONERROR);
		return 0;
	}
	ShowWindow(lpapp->hwnd_MainWindow,SW_SHOW);				//Display the window.
	UpdateWindow(lpapp->hwnd_MainWindow);					//Send WM_PAINT.
	//Messages Loop
	while (GetMessage(&msg,NULL,0,0))	//Get Message if any.
	{
		TranslateMessage(&msg);			//Translate the virtuel keys event.
		DispatchMessage(&msg);			//DispatchMessage to the related window.
	}

	if (hr_coinit==S_OK) CoUninitialize();
	return (int)(msg.wParam);
}



